///////////////////////////////////////////////////////////////////////////////////////////
//
// System iNtrusion Analysis and Reporting Environment (SNARE) - Audit Module
//
// This is the core of the SNARE audit subsystem. The audit module intercepts
// system calls, and saves off details associated with the call to a circular memory
// buffer, which is then read by the audit daemon.
//
///////////////////////////////////////////////////////////////////////////////////////////
// Copyright 1999-2001 InterSect Alliance Pty Ltd - http://www.intersectalliance.com/
///////////////////////////////////////////////////////////////////////////////////////////

// Includes
#include <linux/module.h>
#include <linux/init.h>

#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/time.h>

#include <linux/config.h>
#include <linux/version.h>
#include <linux/reboot.h>
#include <linux/proc_fs.h>
#include <linux/time.h>
#include <linux/tqueue.h>

#include <asm/uaccess.h>
#include <asm/unistd.h>

#include <asm/io.h>
#include <asm/current.h>
#include <asm/uaccess.h>
#include <asm/bitops.h>
#include <asm/unistd.h>

#include <sys/syscall.h>

#include "audit.h"
#include "auditmodule.h"

#include <stdio.h>


// The following line is for kernel 2.4 and above. 2.2 already has a MAP_NR.
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,4,0)
#define MAP_NR(addr) (((unsigned long)(addr)-PAGE_OFFSET) >>PAGE_SHIFT)
#endif

// Global static variables
static char *	buffer_page;			// High level buffer page
static char *	storage_buffer_start;	// The buffer
static int		buf_size_order;			// Buffer size order
static unsigned int buffer_size;		// Buffer sizes
static char*	write_position;			// Current write position in the buffer
static char*	read_position;			// Current read position in the buffer
static int		queue_bytes_free=0;		// Variable used for blocking

// /proc/auditinfo read position.
static char *	info_read_position;		// read position in the information proc entry.

static spinlock_t			spin_lock;					// kernel locking
static struct task_struct *	auditdaemon_task_struct;	// Daemon interaction
unsigned long int flags; 								// CPU flags when using spin_lock_irqsave

extern void *sys_call_table[];

static int AUDIT_IS_RUNNING=0;	// Running flag

// For the /proc/audit device
static short audit_device_open=0;
static short audit_info_open=0;

// How many packets have we lost this session due to low buffer size?
static int lost_events=0;
static int total_events=0;

// Create an array to store the active events
static short active_audit[MAXAUDIT];

struct proc_dir_entry *ProcEntry;
struct proc_dir_entry *ProcInfoEntry;

DECLARE_WAIT_QUEUE_HEAD(audit_queue);
DECLARE_WAIT_QUEUE_HEAD(wait_queue);

// The task queue structure for this task, from tqueue.h
static struct tq_struct Task =
{
//  NULL,    			// Next item in list - queue_task will do this for us
//  0,        			// A flag meaning we haven't been inserted into a task queue yet
  routine: timecheck,	// The function to run
  data: NULL			// The void* parameter for that function
};


// Delivery mechanism - 0 = normal, 1 = with 5 second timeout, 2 = force c2.
static short audit_delivery=0;

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
// Set up file operations that we are capable of handling
// kernel 2.2
static struct file_operations file_ops =
{
	NULL,					// llseek
	auditmodule_read,		// read
	NULL,					// write
	NULL,					// readdir
	NULL,					// poll
	auditmodule_ioctl,		// ioctl
	NULL,					// mmap
	auditmodule_open,		// open
	NULL,					// flush
	auditmodule_close,		// release
	NULL,					// fsync
	NULL,					// fasync
	NULL,					// check_media_change
	NULL,					// revalidate
	NULL					// lock
};

static struct file_operations info_ops =
{
	NULL,					// llseek
	info_read,				// read
	NULL,					// write
	NULL,					// readdir
	NULL,					// poll
	NULL,					// ioctl
	NULL,					// mmap
	NULL,					// open
	NULL,					// flush
	NULL,					// release
	NULL,					// fsync
	NULL,					// fasync
	NULL,					// check_media_change
	NULL,					// revalidate
	NULL					// lock
};
#else
// Kernel 2.4
static struct file_operations file_ops =
{
	read:	auditmodule_read,		// read
	ioctl:	auditmodule_ioctl,		// ioctl
	open:	auditmodule_open,		// open
	release:	auditmodule_close,	// release
};

static struct file_operations info_ops =
{
	read:	info_read,				// read
};
#endif

MODULE_AUTHOR("InterSect Alliance - www.intersectalliance.com");
MODULE_DESCRIPTION("Linux Audit Module");

const long buffersize;			// Startup parameter
MODULE_PARM(buffersize,"l");	// buffersize registered as an long int module parameter.
MODULE_PARM_DESC(buffersize,"Requested size of buffer, in bytes. Optimal values are exponents of PAGE_SIZE [4096 bytes]");

// Parameters for event selection.
const int file_class;			MODULE_PARM(file_class,"i");
MODULE_PARM_DESC(file_class,"File-related audit events");
int a_open;				MODULE_PARM(a_open,"i");
int a_creat;			MODULE_PARM(a_creat,"i");
int a_mkdir;			MODULE_PARM(a_mkdir,"i");
int a_unlink;			MODULE_PARM(a_unlink,"i");
int a_mknod;			MODULE_PARM(a_mknod,"i");
int a_rmdir;			MODULE_PARM(a_rmdir,"i");
int a_symlink;			MODULE_PARM(a_symlink,"i");
int a_link;				MODULE_PARM(a_link,"i");
int a_rename;			MODULE_PARM(a_rename,"i");
int a_truncate;			MODULE_PARM(a_truncate,"i");
//const int a_ftruncate64;	MODULE_PARM(a_ftruncate64,"i");
int a_truncate64;		MODULE_PARM(a_truncate64,"i");
//const int a_ftruncate;		MODULE_PARM(a_ftruncate,"i");

const int file_control_class;	MODULE_PARM(file_control_class,"i");
MODULE_PARM_DESC(file_control_class,"File control related audit events");
int a_chown;			MODULE_PARM(a_chown,"i");
int a_lchown;			MODULE_PARM(a_lchown,"i");
int a_chown32;			MODULE_PARM(a_chown32,"i");
int a_lchown32; 		MODULE_PARM(a_lchown32,"i");
int a_chmod; 			MODULE_PARM(a_chmod,"i");
//const int a_fchmod; 		MODULE_PARM(a_fchmod,"i");
//const int a_fchown;			MODULE_PARM(a_fchown,"i");

const int execute_class;		MODULE_PARM(execute_class,"i");
MODULE_PARM_DESC(execute_class,"Execute related audit events");
int a_execve;			MODULE_PARM(a_execve,"i");
int a_setuid; 			MODULE_PARM(a_setuid,"i");
int a_setreuid; 		MODULE_PARM(a_setreuid,"i");
int a_setresuid; 		MODULE_PARM(a_setresuid,"i");
int a_setuid32; 		MODULE_PARM(a_setuid32,"i");
int a_setreuid32; 		MODULE_PARM(a_setreuid32,"i");
int a_setresuid32;	 	MODULE_PARM(a_setresuid32,"i");
int a_setgid; 			MODULE_PARM(a_setgid,"i");
int a_setregid; 		MODULE_PARM(a_setregid,"i");
int a_setresgid;		MODULE_PARM(a_setresgid,"i");
int a_setgid32;			MODULE_PARM(a_setgid32,"i");
int a_setregid32;	 	MODULE_PARM(a_setregid32,"i");
int a_setresgid32;		MODULE_PARM(a_setresgid32,"i");
int a_exit;				MODULE_PARM(a_exit,"i");

const int admin_class;		MODULE_PARM(admin_class,"i");
MODULE_PARM_DESC(admin_class,"Administrative audit events");
int a_reboot;			MODULE_PARM(a_reboot,"i");
int a_chroot;			MODULE_PARM(a_chroot,"i");

int all;				MODULE_PARM(all,"i");
MODULE_PARM_DESC(all,"Turn on all audit events.");




////////////////////////////////////////////////////////////////////////////////
// Initialise the module
// Set up our event overlay functions.
////////////////////////////////////////////////////////////////////////////////
int init_module(void)
{
	long newbuffersize=0;
	int count=0;
	
	AUDIT_IS_RUNNING = FALSE;
	auditdaemon_task_struct=NULL;

	// Initialise our active event storage array.
	for(count=0;count <= MAXAUDIT; count++) {
		active_audit[count]=0;
	}

	// Kernel lock
	spin_lock = SPIN_LOCK_UNLOCKED;

	if(buffersize)
	{
		if(buffersize < AUDITMODULE_MIN_BUF) {
			printk("Audit: Buffer size must be a minimum %ld - setting to that value.\n",AUDITMODULE_MIN_BUF);
			newbuffersize = AUDITMODULE_MIN_BUF;
		}
		// printk("Audit: Buffer size passed as a parameter to the module: %ld\n",buffersize);
	} else {
		newbuffersize=AUDITMODULE_DEFAULT_BUF;
	}

	// Allocate a buffer
	// Since our memory requirements are too large for kmalloc to handle
	// use our own memory grabbing routine to catch a single large memory page.
	
	if(get_buffer(newbuffersize))
	{
		printk("Audit: Could not allocate buffers for audit data\n");
		return(1);
	}

	if(storage_buffer_start == NULL)
	{
		printk("Audit: Could not allocate buffers for audit data\n");
		return(1);
	}

	// printk("Audit: Created and allocated buffers of %ld bytes\n",newbuffersize);
	
	// From here on in, the read routine could be signaled.
	// Therefore, wherever we write to the audit buffer, or we rely on
	// any of the variables that the read routine changes,
	// we should lock the kernel first.

	// Would do this using proc_register in kernel 2.2 - see http://linuxkernel.to/module/port-2.4/eng/lkp-all.html
	if ((ProcEntry = create_proc_entry(AUDITDEV_NAME, S_IRUGO | S_IWUSR, NULL)) == NULL)
		return(1);
		
	ProcEntry->proc_fops = &file_ops;

	if ((ProcInfoEntry = create_proc_entry(AUDITINFO_NAME, S_IRUGO | S_IWUSR, NULL)) == NULL)
		return(1);
		
	ProcInfoEntry->proc_fops = &info_ops;


	// Now: Verify what audit events the user wants, and insert ourselves into the
	// relevant system calls.

	printk("SNARE Audit Module: Version %d.%d%d initialized\n", AUDITMODULE_MAJOR_VERSION,
			AUDITMODULE_MINOR_VERSION, AUDITMODULE_PATCH_VERSION);

	// Set up auditable functions
	if(a_open || all || file_class)					audit_on(SYS_open);
	if(a_creat || all || file_class)				audit_on(SYS_creat);
	if(a_execve || all || execute_class)			audit_on(SYS_execve);
	if(a_exit || all || execute_class)				audit_on(SYS_exit);
	if(a_mkdir || all || file_class)				audit_on(SYS_mkdir);
	if(a_unlink || all || file_class)				audit_on(SYS_unlink);
	if(a_mknod || all || file_class)				audit_on(SYS_mknod);
	if(a_rmdir || all || file_class)				audit_on(SYS_rmdir);
	if(a_chown || all || file_control_class)		audit_on(SYS_chown);
	if(a_lchown || all || file_control_class)		audit_on(SYS_lchown);
	// Why are these two not picked up from asm/unistd.h under kernel 2.2?
	//#define __NR_chown32		212
	//#define __NR_lchown32		198 
	if(a_chown32 || all || file_control_class)		audit_on(SYS_chown32);
	if(a_lchown32 || all || file_control_class)		audit_on(SYS_lchown32);
	if(a_chmod || all || file_control_class)		audit_on(SYS_chmod);
	if(a_symlink || all || file_class)				audit_on(SYS_symlink);
	if(a_link || all || file_class)					audit_on(SYS_link);
	if(a_rename || all || file_class)				audit_on(SYS_rename);
	if(a_reboot || all || admin_class)				audit_on(SYS_reboot);
	if(a_truncate || all || file_class)				audit_on(SYS_truncate);
	if(a_chroot || all || admin_class)				audit_on(SYS_chroot);
	if(a_setuid || all || execute_class)			audit_on(SYS_setuid);
	if(a_setreuid || all || execute_class)			audit_on(SYS_setreuid);
	if(a_setresuid || all || execute_class)			audit_on(SYS_setresuid);
	if(a_setuid32 || all || execute_class)			audit_on(SYS_setuid32);
	if(a_setreuid32 || all || execute_class)		audit_on(SYS_setreuid32);
	if(a_setresuid32 || all || execute_class)		audit_on(SYS_setresuid32);
	if(a_setgid || all || execute_class)			audit_on(SYS_setgid);
	if(a_setregid || all || execute_class)			audit_on(SYS_setregid);
	if(a_setresgid || all || execute_class)			audit_on(SYS_setresgid);
	if(a_setgid32 || all || execute_class)			audit_on(SYS_setgid32);
	if(a_setregid32 || all || execute_class)		audit_on(SYS_setregid32);
	if(a_setresgid32 || all || execute_class)		audit_on(SYS_setresgid32);
//	if(a_ftruncate || all || file_class)			audit_on(SYS_ftruncate);	// Not required - truncate covers this.
//	if(a_fchmod || all || file_control_class)		audit_on(SYS_fchmod);
//	if(a_ftruncate64 || all || file_class)			audit_on(SYS_ftruncate64);
	if(a_truncate64 || all || file_class)			audit_on(SYS_truncate64);
//	if(a_fchown || all || file_control_class)		audit_on(SYS_fchown);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
// This routine handles ioctl messages from the audit daemon.
////////////////////////////////////////////////////////////////////////////////

int auditmodule_ioctl(struct inode* node, struct file*  the_file, unsigned int  command, unsigned long arg)
{
	static int counter=0;

	// If the audit module is running, there is not much point starting again!
	if((AUDIT_IS_RUNNING == TRUE) && (command == AUDIT_START))
		return -EBUSY;

	if(command == AUDIT_START) {
		AUDIT_IS_RUNNING = TRUE;
	} else if(command == AUDIT_STOP) {
		AUDIT_IS_RUNNING = FALSE;
	} else if(command == AUDIT_LOSTEVENTS) {
		return(lost_events);
	} else if(command == AUDIT_TOTALEVENTS) {
		return(total_events);
	} else if(command == AUDIT_INFO) {
		// This is designed (with the assistance of the static int at the top of this routine)
		// to return one system call number per AUDIT_EVENTLIST ioctl.
		// the audit daemon code will look something like this:
		// while(event=ioctl(/dev/audit AUDIT_EVENTLIST)) { print the name of the event }
		// If we've hit the end of the array, reset our counter, and return 'end of array' (0)
		//
		// Note that we don't use this capability within auditd at the moment.
		//
		counter++;

		while(active_audit[counter]==0 && counter <= MAXAUDIT) {
			counter++;
		}
		
		if(counter <= MAXAUDIT) {
			return(active_audit[counter]);
		} else {
			counter=0;
			return(0);
		}
	} else if (command == AUDIT_EVENT_ON) {
		audit_on(arg);
	} else if (command == AUDIT_EVENT_OFF) {
		audit_off(arg);
		// Eg: ioctl(x,SYS_exit);
	} else if (command == AUDIT_FLUSH) {
			int counter;
			for(counter=1; counter<=MAXAUDIT; counter++) {
				if(active_audit[counter]) audit_off(active_audit[counter]);
			}		

			// Also reset the lost/total event counters here.
			lost_events=0;
			total_events=0;
	} else if ( command == AUDIT_DELIVERY) {
		// arg: 0 - normal, non-forced delivery
		// arg: 1 - attempt to deliver, within timeout period.
		// arg: 2 - force delivery. System performance gradually degrades as process stack fills.
		audit_delivery=(int)arg;
	} else {
		return -ENOSYS;
	}

	return 0;
}


////////////////////////////////////////////////////////////////////////////////
// Here we give the user the data.
// NOTE that this module may be executed at any time - even in the middle
// of the execution of something else. Hence, any variables that are changed
// by this routine should be surrounded by a kernel lock if they are used
// elsewhere!
//
// In particular:
//	write_position
//  read_position
//
// Let me just repeat this in another way:
// WARNING: DO NOT CHANGE OR RELY ON THE VARIABLES ABOVE UNLESS YOU ARE
//		  IN A SPIN_LOCKED STATE - THEY MAY BE CORRUPTED BY THE
//		  AUDIT_READ MODULE, WHICH IS CALLED BY SIGNAL/INTERRUPT
//
// Unfortunately, we also need to spin-lock within this routine also.
// Otherwise, an interrupt might occur when we were half way through modifying
// something like write_position - and if the interrupt happens to generate an
// audit event, then we may be copying data to the wrong location in the array,
// leading to data corruption.
// Spin locking too much code slows down the system unfortunately. I'll try
// and keep it minimal.
////////////////////////////////////////////////////////////////////////////////
// Note: The following routine uses a rotating (tail-chaser) buffer.
//
// Examples: (. = free, # = contains data yet to be read.)
//
// 100	 110				 150						200
// [.......######################..........................]
//	   read				  write
//
// 100	 110				 150						200
// [########.....................##########################]
//	   write				 read
//

static ssize_t auditmodule_read(struct file *file, char *filebuffer, size_t length, loff_t *ppos)
{
	// Number of bytes actually written to the buffer
	int bytes_read = 0;

	int bytes_avail_to_send=0;
	int bytes_to_write=0;
	int contiguous_bytes_avail=0;

	// NOTE: WILL NEED TO LOCK HERE
	// because both read_position and write_position may potentially be modified by other routines.

	spin_lock_irqsave(&spin_lock, flags);

	// Is there data to be read?
	// remember: read position can catch up to the write position,
	// but write position can never catch the read position!
	// Therefore, if read_position == write_position, we can be confident that
	// there is no data in the buffer.
	
	if(read_position == write_position)
	{
		spin_unlock_irqrestore(&spin_lock, flags);
		// Nope. No data.		
		return(0);
	}

	// Ok, we have some data available to write.
	// Since we spin lock our copies to the buffer, it's safe to say that if data
	// is available, then we have a full audit record in the buffer (if not many).

	// Work out the number of bytes available to be read.
	// Difference between read and write buffers (rotating ring buffer)

	if(write_position > read_position)
	{
		// Work out the number of bytes available to be written to the user.
		bytes_avail_to_send=write_position - read_position;
		contiguous_bytes_avail=bytes_avail_to_send; // Since we have a single contiguous block of data.
	} else {
		// Add the distance between the read position and the end of the buffer to
		// the distance between the start of the buffer and the write position.
		bytes_avail_to_send=((storage_buffer_start + buffer_size) - read_position) + (write_position - storage_buffer_start);
		contiguous_bytes_avail=(storage_buffer_start + buffer_size) - read_position;
	}

	if(bytes_avail_to_send < length)
	{
		bytes_to_write=bytes_avail_to_send;
	} else {
		bytes_to_write=length;
	}
	// Save the number of bytes off for our return code.
	bytes_read=bytes_to_write;

		
	// copy to the user the lesser of:
	//  bytes requested, or
	//  bytes available
	
	// Write out the data prior to the end of the buffer
	// Is the amount of data requested/available contiguous in the buffer?
	if(bytes_to_write <= contiguous_bytes_avail)
	{
		// This will always be the case if read position < write position.
		// But MAY not be the case of write position < read position
		// copy this data to userspace.
		if(copy_to_user(filebuffer,read_position,bytes_to_write))
		{
			spin_unlock_irqrestore(&spin_lock, flags);
			return -EFAULT;
		}

		read_position+=bytes_to_write;
		if(read_position >= (storage_buffer_start+buffer_size))
		{
			// We are at the end of the buffer. Rotate.
			read_position=storage_buffer_start;
		}
	} else {
		if(copy_to_user(filebuffer,read_position,contiguous_bytes_avail))
		{
			spin_unlock_irqrestore(&spin_lock, flags);
				return -EFAULT;
		}

		// Reset our read position to the start of the buffer.
		read_position=storage_buffer_start;
		// Subtract the number of bytes we have already written
		bytes_to_write -= contiguous_bytes_avail;

		if(copy_to_user(filebuffer+contiguous_bytes_avail,read_position,bytes_to_write))
		{
			spin_unlock_irqrestore(&spin_lock, flags);
			return -EFAULT;
		}

		// Set our read position to the appropriate location.
		read_position+=bytes_to_write;
	}

	spin_unlock_irqrestore(&spin_lock, flags);

	if(audit_delivery == DELIVERY_TIMEOUT || audit_delivery == DELIVERY_C2) {
		// Signal the wait queue to continue processing.
		wake_up(&audit_queue);
	}

   	// Read functions are supposed to return the number of bytes actually inserted into the buffer
	return bytes_read;
}


////////////////////////////////////////////////////////////////////////////////
// open the device
////////////////////////////////////////////////////////////////////////////////
static int auditmodule_open(struct inode* node, struct file*  the_file)
{
	int device_minor = MINOR(node->i_rdev) & 0xf;  // Device minor number

	// reset the lost packets indicator
	lost_events=0;
	total_events=0;

	// Can only have one audit device, and it can only be open once
	if(device_minor)
		return -ENODEV;
	if(audit_device_open)
		return -EBUSY;

	// Ok, we're open.
	audit_device_open=1;

	// Fetch the task structure of the process that opened the device
	auditdaemon_task_struct = current;

	printk("Auditing: Process %d has opened the device\n", auditdaemon_task_struct->pid);

	MOD_INC_USE_COUNT; // Increment module count

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
// close the device and release resources
////////////////////////////////////////////////////////////////////////////////
static int auditmodule_close(struct inode* node, struct file*  the_file)
{
	auditdaemon_task_struct = NULL;
	AUDIT_IS_RUNNING = FALSE;
	audit_device_open=0;
	auditdaemon_task_struct=NULL;

	printk("AUDITMODULE: Audit daemon has closed dev/audit.\n");
	printk("AUDITMODULE: Events lost due low buffer size this session: %d\n",lost_events);
	printk("AUDITMODULE: Total Events processed this session: %d\n",total_events);
	lost_events=0;
	total_events=0;
	
	// Reset all our buffers here.
	// con: reset buffers may lose us some events.
	// pro: we could have been terminated half way through reading something, and auditd will not recover.
	spin_lock_irqsave(&spin_lock, flags);
	write_position=read_position=storage_buffer_start;
	spin_unlock_irqrestore(&spin_lock, flags);

	MOD_DEC_USE_COUNT; // Decrement module usage
	
	return 0;
}

// /proc/auditinfo open.
static int info_open(struct inode* node, struct file*  the_file)
{
	int device_minor = MINOR(node->i_rdev) & 0xf;  /* Device minor number */

	// Only one open at a time thanks.
	if(device_minor)
		return -ENODEV;
	if(audit_info_open)
		return -EBUSY;

	// Ok, we're open.
	audit_info_open=1;
	info_read_position=(char *)NULL;

	return 0;
}

static int info_close(struct inode* node, struct file*  the_file)
{
	audit_info_open=0;
	info_read_position=(char *)NULL;

	return 0;
}

static ssize_t info_read(struct file *file, char *filebuffer, size_t length, loff_t *ppos)
{
	// Would prefer to dynamically malloc this, but we run the risk of not freeing at program termination.
	static char buffer[150]="";
	static int finished=0;
	int bytes_to_write=0;
	
	if(finished) {
		finished=0;
		// Return zero to indicate EOF
		return(0);
	}

	if(info_read_position == (char *)NULL) {
		char processid[20]="No Daemon Running";
				
		if(auditdaemon_task_struct) {
			if(auditdaemon_task_struct->pid) {
				// WOULD REALLY prefer snprintf here.
				sprintf(processid,"%d",(int)auditdaemon_task_struct->pid);
			}
		}
		
		// Would prefer to use snprintf here.
		// However, the size of these variables are under our control
		// (except pid, which is going to be <= maxint. (5 characters)
		// Therefore we need a buffer max of around 100 characters. Allow 150 just in case.
		// SNARE Version: 999.999.999XAudit Active: 9XAudit Process ID: 99999XEvents Lost This Session: 99999XX
		sprintf(buffer,"SNARE Version: %d.%d.%d\nAudit Active: %d\nAudit Process ID: %s\nEvents Lost This Session: %d\nEvents Processed This Session: %d\n",
			AUDITMODULE_MAJOR_VERSION,AUDITMODULE_MINOR_VERSION,AUDITMODULE_PATCH_VERSION,
			AUDIT_IS_RUNNING,
			processid, lost_events,total_events);
		
		info_read_position=buffer;
	}
			
	bytes_to_write=strlen(info_read_position);
	if(bytes_to_write > length) {
		if(copy_to_user(filebuffer,info_read_position,length))
		{
			return -EFAULT;
		}
		info_read_position+=length;
		return(length);
	} else {
		// Excellent The entire buffer has been completed.
		// Free the memory, return the pointer to zero.
		if(copy_to_user(filebuffer,info_read_position,bytes_to_write))
		{
			return -EFAULT;
		}
		info_read_position=(char *)NULL;
		finished=1;
		return(bytes_to_write);
	}

	return(0);
}

////////////////////////////////////////////////////////////////////////////////
// Allocate read/write buffer
//
// Since we're only called once per module invocation, no need to check to see if
// memory has already been allocated.
////////////////////////////////////////////////////////////////////////////////

int get_buffer(unsigned int buffer_request)
{
	static char* write_buf_end;
	static char* buf_writelimit;

	int total_size;  // total amount to be allocated
	int total_pages; // # of pages this equals */
	unsigned int temp_start, temp_end; // temp page #'s to set/unset reserved bit
	
	printk("Audit: Allocating buffer space...");
	// check the buf size - if not in range, use the default and keep going
	// either way, set the size to be allocated
	buffer_size = buffer_request;
	total_size=buffer_request;

	// How many pages do we need?
	total_pages = total_size / PAGE_SIZE;
	if((buffer_request % PAGE_SIZE) != 0)
		total_pages += 1;
  
	// Find the nearest power of 2 and allocate the pages with it
	buf_size_order = 0;
	while((total_pages = total_pages >> 1) != 0)
	{
		// Is this an odd number greater than 2
		if((total_pages > 2) && ((total_pages % 2) != 0))
			total_pages += 1;
		buf_size_order += 1;
 	}

	// printk("Audit: Buffer size order is %d\n", buf_size_order);

	buffer_page = (char*) __get_free_pages(GFP_ATOMIC, buf_size_order);
	if(buffer_page == NULL) // not enough memory
		return -ENOMEM;

	// Set the start and end of buffering region
	temp_start = (unsigned int) buffer_page;
	temp_end   = (unsigned int) buffer_page + total_size;

	// Should probably rewrite to use virt_to_page.
	// Find another routine that allocates an arbitrary sized page.

	// For each page, set the reserved bit (allows userspace to access them)
	for(; temp_start < temp_end; temp_start += PAGE_SIZE)
		set_bit(PG_reserved, &(mem_map[MAP_NR(temp_start)].flags));

	// Set up all the buffer pointers
	storage_buffer_start = buffer_page;

	write_buf_end = storage_buffer_start + buffer_size;
	buf_writelimit = write_buf_end;

	write_position = storage_buffer_start;
	read_position = storage_buffer_start;
	
	// printk("Audit: Allocated buffers of %d bytes\n", buffer_size);

	return 0;
}

////////////////////////////////////////////////////////////////////////////////
// Free buffers. Give the pages back to the kernel.
////////////////////////////////////////////////////////////////////////////////

int free_buffer(unsigned int buffer_request)
{
	unsigned int temp_start, temp_end; // temp page #'s to set/unset reserved bit

	// free buffers and pages, and exit
	if(buffer_page != NULL)
	{
		/* Set the start and end of buffering region */
		temp_start = (uint32_t) buffer_page;
		temp_end   = (uint32_t) buffer_page + buffer_request;
		// printk("Auditing: Freeing the allocated buffer\n");
		// Unset the reserved bit
		for(; temp_start < temp_end; temp_start += PAGE_SIZE)
			clear_bit(PG_reserved, &(mem_map[MAP_NR(temp_start)].flags));
		// printk("Auditing: cleaning...\n");
		// Give back the pages
		free_pages((uint32_t) buffer_page, buf_size_order);
		return(1);
	} else {
		return(0); // Problem.. mem not found to be allocated?
	}
}


////////////////////////////////////////////////////////////////////////////////
// Signal the audit daemon that is reading from /dev/audit (if it is there).
// Note that this is called from within audit_event, but is not IRQ locked.
////////////////////////////////////////////////////////////////////////////////
int signal_auditd()
{
	static struct siginfo sig_info;	/* Signal information */

	// Not much use sending signals to a non-existant process.
	if(audit_device_open && AUDIT_IS_RUNNING)
	{
		// Setup signal information and send it
		sig_info.si_signo = SIGIO;
		sig_info.si_errno = 0;
		sig_info.si_code  = SI_KERNEL;
		send_sig_info(SIGIO, &sig_info, auditdaemon_task_struct);
		return(1);
	}
	return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Free buffers, release our overlay functions and exit.
////////////////////////////////////////////////////////////////////////////////
void cleanup_module(void)
{
	int counter=0;
	// printk("Auditing: Cleaning up module\n");
	// remove the proc entry, so the daemon won't try to get more data
	remove_proc_entry(AUDITDEV_NAME,NULL);
	remove_proc_entry(AUDITINFO_NAME,NULL);

	// Release the audited functions - skip AUDIT_NULL though.
	for(counter=1; counter<=MAXAUDIT; counter++) {
		if(active_audit[counter]) audit_off(active_audit[counter]);
	}

	// Free our buffers
	free_buffer(buffer_size);

	printk("SNARE Audit Module: Exiting\n");
	buffer_page = NULL; /* just to be sure... */
}

// Don't turn it on twice.
void audit_on(int auditnum)
{
	// printk("Debug: Turning on event %d\n",auditnum);
	switch (auditnum)
	{
		case SYS_open:
			if(active_audit[AUDIT_open]) return;
			orig_open=sys_call_table[SYS_open];
			sys_call_table[SYS_open]=audit_open;
			active_audit[AUDIT_open]=SYS_open;
			break;
		case SYS_creat:
			if(active_audit[AUDIT_creat]) return;
			orig_creat=sys_call_table[SYS_creat];
			sys_call_table[SYS_creat]=audit_creat;
			active_audit[AUDIT_creat]=SYS_creat;
			break;
		case SYS_execve:
			if(active_audit[AUDIT_execve]) return;
			orig_execve=sys_call_table[SYS_execve];
			sys_call_table[SYS_execve]=audit_execve;
			active_audit[AUDIT_execve]=SYS_execve;
			break;
		case SYS_exit:
			if(active_audit[AUDIT_exit]) return;
			orig_exit=sys_call_table[SYS_exit];
			sys_call_table[SYS_exit]=audit_exit;
			active_audit[AUDIT_exit]=SYS_exit;
			break;
		case SYS_mkdir:
			if(active_audit[AUDIT_mkdir]) return;
			orig_mkdir=sys_call_table[SYS_mkdir];
			sys_call_table[SYS_mkdir]=audit_mkdir;
			active_audit[AUDIT_mkdir]=SYS_mkdir;
			break;
		case SYS_unlink:
			if(active_audit[AUDIT_unlink]) return;
			orig_unlink=sys_call_table[SYS_unlink];
			sys_call_table[SYS_unlink]=audit_unlink;
			active_audit[AUDIT_unlink]=SYS_unlink;
			break;
		case SYS_mknod:
			if(active_audit[AUDIT_mknod]) return;
			orig_mknod=sys_call_table[SYS_mknod];
			sys_call_table[SYS_mknod]=audit_mknod;
			active_audit[AUDIT_mknod]=SYS_mknod;
			break;
		case SYS_rmdir:
			if(active_audit[AUDIT_rmdir]) return;
			orig_rmdir=sys_call_table[SYS_rmdir];
			sys_call_table[SYS_rmdir]=audit_rmdir;
			active_audit[AUDIT_rmdir]=SYS_rmdir;
			break;
		case SYS_chown:
			if(active_audit[AUDIT_chown]) return;
			orig_chown=sys_call_table[SYS_chown];
			sys_call_table[SYS_chown]=audit_chown;
			active_audit[AUDIT_chown]=SYS_chown;
			break;
		case SYS_lchown:
			if(active_audit[AUDIT_lchown]) return;
			orig_lchown=sys_call_table[SYS_lchown];
			sys_call_table[SYS_lchown]=audit_lchown;
			active_audit[AUDIT_lchown]=SYS_lchown;
			break;
		case SYS_chown32:
			if(active_audit[AUDIT_chown32]) return;
			orig_chown32=sys_call_table[SYS_chown32];
			sys_call_table[SYS_chown32]=audit_chown32;
			active_audit[AUDIT_chown32]=SYS_chown32;
			break;
		case SYS_lchown32:
			if(active_audit[AUDIT_lchown32]) return;
			orig_lchown32=sys_call_table[SYS_lchown32];
			sys_call_table[SYS_lchown32]=audit_lchown32;
			active_audit[AUDIT_lchown32]=SYS_lchown32;
			break;
		case SYS_chmod:
			if(active_audit[AUDIT_chmod]) return;
			orig_chmod=sys_call_table[SYS_chmod];
			sys_call_table[SYS_chmod]=audit_chmod;
			active_audit[AUDIT_chmod]=SYS_chmod;
			break;
		case SYS_symlink:
			if(active_audit[AUDIT_symlink]) return;
			orig_symlink=sys_call_table[SYS_symlink];
			sys_call_table[SYS_symlink]=audit_symlink;
			active_audit[AUDIT_symlink]=SYS_symlink;
			break;
		case SYS_link:
			if(active_audit[AUDIT_link]) return;
			orig_link=sys_call_table[SYS_link];
			sys_call_table[SYS_link]=audit_link;
			active_audit[AUDIT_link]=SYS_link;
			break;
		case SYS_rename:
			if(active_audit[AUDIT_rename]) return;
			orig_rename=sys_call_table[SYS_rename];
			sys_call_table[SYS_rename]=audit_rename;
			active_audit[AUDIT_rename]=SYS_rename;
			break;
		case SYS_reboot:
			if(active_audit[AUDIT_reboot]) return;
			orig_reboot=sys_call_table[SYS_reboot];
			sys_call_table[SYS_reboot]=audit_reboot;
			active_audit[AUDIT_reboot]=SYS_reboot;
			break;
		case SYS_truncate:
			if(active_audit[AUDIT_truncate]) return;
			orig_truncate=sys_call_table[SYS_truncate];
			sys_call_table[SYS_truncate]=audit_truncate;
			active_audit[AUDIT_truncate]=SYS_truncate;
			break;
		case SYS_chroot:
			if(active_audit[AUDIT_chroot]) return;
			orig_chroot=sys_call_table[SYS_chroot];
			sys_call_table[SYS_chroot]=audit_chroot;
			active_audit[AUDIT_chroot]=SYS_chroot;
			break;
		case SYS_setuid:
			if(active_audit[AUDIT_setuid]) return;
			orig_setuid=sys_call_table[SYS_setuid];
			sys_call_table[SYS_setuid]=audit_setuid;
			active_audit[AUDIT_setuid]=SYS_setuid;
			break;
		case SYS_setreuid:
			if(active_audit[AUDIT_setreuid]) return;
			orig_setreuid=sys_call_table[SYS_setreuid];
			sys_call_table[SYS_setreuid]=audit_setreuid;
			active_audit[AUDIT_setreuid]=SYS_setreuid;
			break;
		case SYS_setresuid:
			if(active_audit[AUDIT_setresuid]) return;
			orig_setresuid=sys_call_table[SYS_setresuid];
			sys_call_table[SYS_setresuid]=audit_setresuid;
			active_audit[AUDIT_setresuid]=SYS_setresuid;
			break;
		case SYS_setuid32:
			if(active_audit[AUDIT_setuid32]) return;
			orig_setuid32=sys_call_table[SYS_setuid32];
			sys_call_table[SYS_setuid32]=audit_setuid32;
			active_audit[AUDIT_setuid32]=SYS_setuid32;
			break;
		case SYS_setreuid32:
			if(active_audit[AUDIT_setreuid32]) return;
			orig_setreuid32=sys_call_table[SYS_setreuid32];
			sys_call_table[SYS_setreuid32]=audit_setreuid32;
			active_audit[AUDIT_setreuid32]=SYS_setreuid32;
			break;
		case SYS_setresuid32:
			if(active_audit[AUDIT_setresuid32]) return;
			orig_setresuid32=sys_call_table[SYS_setresuid32];
			sys_call_table[SYS_setresuid32]=audit_setresuid32;
			active_audit[AUDIT_setresuid32]=SYS_setresuid32;
			break;
		case SYS_setgid:
			if(active_audit[AUDIT_setgid]) return;
			orig_setgid=sys_call_table[SYS_setgid];
			sys_call_table[SYS_setgid]=audit_setgid;
			active_audit[AUDIT_setgid]=SYS_setgid;
			break;
		case SYS_setregid:
			if(active_audit[AUDIT_setregid]) return;
			orig_setregid=sys_call_table[SYS_setregid];
			sys_call_table[SYS_setregid]=audit_setregid;
			active_audit[AUDIT_setregid]=SYS_setregid;
			break;
		case SYS_setresgid:
			if(active_audit[AUDIT_setresgid]) return;
			orig_setresgid=sys_call_table[SYS_setresgid];
			sys_call_table[SYS_setresgid]=audit_setresgid;
			active_audit[AUDIT_setresgid]=SYS_setresgid;
			break;
		case SYS_setgid32:
			if(active_audit[AUDIT_setgid32]) return;
			orig_setgid32=sys_call_table[SYS_setgid32];
			sys_call_table[SYS_setgid32]=audit_setgid32;
			active_audit[AUDIT_setgid32]=SYS_setgid32;
			break;
		case SYS_setregid32:
			if(active_audit[AUDIT_setregid32]) return;
			orig_setregid32=sys_call_table[SYS_setregid32];
			sys_call_table[SYS_setregid32]=audit_setregid32;
			active_audit[AUDIT_setregid32]=SYS_setregid32;
			break;
		case SYS_setresgid32:
			if(active_audit[AUDIT_setresgid32]) return;
			orig_setresgid32=sys_call_table[SYS_setresgid32];
			sys_call_table[SYS_setresgid32]=audit_setresgid32;
			active_audit[AUDIT_setresgid32]=SYS_setresgid32;
			break;
		case SYS_truncate64:
			if(active_audit[AUDIT_truncate64]) return;
			orig_truncate64=sys_call_table[SYS_truncate64];
			sys_call_table[SYS_truncate64]=audit_truncate64;
			active_audit[AUDIT_truncate64]=SYS_truncate64;
			break;
// No need for the calls below, they are covered by their non f* calls (eg: ftruncate calls truncate64 on my system).
//		case SYS_ftruncate:
//			if(active_audit[AUDIT_ftruncate]) return;
//			orig_ftruncate=sys_call_table[SYS_ftruncate];
//			sys_call_table[SYS_ftruncate]=audit_ftruncate;
//			active_audit[AUDIT_ftruncate]=SYS_ftruncate;
//			break;
//		case SYS_fchmod:
//			if(active_audit[AUDIT_fchmod]) return;
//			orig_fchmod=sys_call_table[SYS_fchmod];
//			sys_call_table[SYS_fchmod]=audit_fchmod;
//			active_audit[AUDIT_fchmod]=SYS_fchmod;
//			break;
//		case SYS_ftruncate64:
//			if(active_audit[AUDIT_ftruncate64]) return;
//			orig_ftruncate64=sys_call_table[SYS_ftruncate64];
//			sys_call_table[SYS_ftruncate64]=audit_ftruncate64;
//			active_audit[AUDIT_ftruncate64]=SYS_ftruncate64;
//			break;
//		case SYS_fchown:
//			if(active_audit[AUDIT_fchown]) return;
//			orig_fchown=sys_call_table[SYS_fchown];
//			sys_call_table[SYS_fchown]=audit_fchown;
//			active_audit[AUDIT_fchown]=SYS_fchown;
//			break;
	}
}

void audit_off(int auditnum)
{
	switch (auditnum)
	{
		case SYS_open:
			if(active_audit[AUDIT_open]==0) return;
			sys_call_table[SYS_open]=orig_open;
			active_audit[AUDIT_open]=0;
			break;
		case SYS_creat:
			if(active_audit[AUDIT_creat]==0) return;
			sys_call_table[SYS_creat]=orig_creat;
			active_audit[AUDIT_creat]=0;
			break;
		case SYS_execve:
			if(active_audit[AUDIT_execve]==0) return;
			sys_call_table[SYS_execve]=orig_execve;
			active_audit[AUDIT_execve]=0;
			break;
		case SYS_exit:
			if(active_audit[AUDIT_exit]==0) return;
			sys_call_table[SYS_exit]=orig_exit;
			active_audit[AUDIT_exit]=0;
			break;
		case SYS_mkdir:
			if(active_audit[AUDIT_mkdir]==0) return;
			sys_call_table[SYS_mkdir]=orig_mkdir;
			active_audit[AUDIT_mkdir]=0;
			break;
		case SYS_unlink:
			if(active_audit[AUDIT_unlink]==0) return;
			sys_call_table[SYS_unlink]=orig_unlink;
			active_audit[AUDIT_unlink]=0;
			break;
		case SYS_mknod:
			if(active_audit[AUDIT_mknod]==0) return;
			sys_call_table[SYS_mknod]=orig_mknod;
			active_audit[AUDIT_mknod]=0;
			break;
		case SYS_rmdir:
			if(active_audit[AUDIT_rmdir]==0) return;
			sys_call_table[SYS_rmdir]=orig_rmdir;
			active_audit[AUDIT_rmdir]=0;
			break;
		case SYS_chown:
			if(active_audit[AUDIT_chown]==0) return;
			sys_call_table[SYS_chown]=orig_chown;
			active_audit[AUDIT_chown]=0;
			break;
		case SYS_lchown:
			if(active_audit[AUDIT_lchown]==0) return;
			sys_call_table[SYS_lchown]=orig_lchown;
			active_audit[AUDIT_lchown]=0;
			break;
		case SYS_chown32:
			if(active_audit[AUDIT_chown32]==0) return;
			sys_call_table[SYS_chown32]=orig_chown32;
			active_audit[AUDIT_chown32]=0;
			break;
		case SYS_lchown32:
			if(active_audit[AUDIT_lchown32]==0) return;
			sys_call_table[SYS_lchown32]=orig_lchown32;
			active_audit[AUDIT_lchown32]=0;
			break;
		case SYS_chmod:
			if(active_audit[AUDIT_chmod]==0) return;
			sys_call_table[SYS_chmod]=orig_chmod;
			active_audit[AUDIT_chmod]=0;
			break;
		case SYS_symlink:
			if(active_audit[AUDIT_symlink]==0) return;
			sys_call_table[SYS_symlink]=orig_symlink;
			active_audit[AUDIT_symlink]=0;
			break;
		case SYS_link:
			if(active_audit[AUDIT_link]==0) return;
			sys_call_table[SYS_link]=orig_link;
			active_audit[AUDIT_link]=0;
			break;
		case SYS_rename:
			if(active_audit[AUDIT_rename]==0) return;
			sys_call_table[SYS_rename]=orig_rename;
			active_audit[AUDIT_rename]=0;
			break;
		case SYS_reboot:
			if(active_audit[AUDIT_reboot]==0) return;
			sys_call_table[SYS_reboot]=orig_reboot;
			active_audit[AUDIT_reboot]=0;
			break;
		case SYS_truncate:
			if(active_audit[AUDIT_truncate]==0) return;
			sys_call_table[SYS_truncate]=orig_truncate;
			active_audit[AUDIT_truncate]=0;
			break;
		case SYS_chroot:
			if(active_audit[AUDIT_chroot]==0) return;
			sys_call_table[SYS_chroot]=orig_chroot;
			active_audit[AUDIT_chroot]=0;
			break;
		case SYS_setuid:
			if(active_audit[AUDIT_setuid]==0) return;
			sys_call_table[SYS_setuid]=orig_setuid;
			active_audit[AUDIT_setuid]=0;
			break;
		case SYS_setreuid:
			if(active_audit[AUDIT_setreuid]==0) return;
			sys_call_table[SYS_setreuid]=orig_setreuid;
			active_audit[AUDIT_setreuid]=0;
			break;
		case SYS_setresuid:
			if(active_audit[AUDIT_setresuid]==0) return;
			sys_call_table[SYS_setresuid]=orig_setresuid;
			active_audit[AUDIT_setresuid]=0;
			break;
		case SYS_setuid32:
			if(active_audit[AUDIT_setuid32]==0) return;
			sys_call_table[SYS_setuid32]=orig_setuid32;
			active_audit[AUDIT_setuid32]=0;
			break;
		case SYS_setreuid32:
			if(active_audit[AUDIT_setreuid32]==0) return;
			sys_call_table[SYS_setreuid32]=orig_setreuid32;
			active_audit[AUDIT_setreuid32]=0;
			break;
		case SYS_setresuid32:
			if(active_audit[AUDIT_setresuid32]==0) return;
			sys_call_table[SYS_setresuid32]=orig_setresuid32;
			active_audit[AUDIT_setresuid32]=0;
			break;
		case SYS_setgid:
			if(active_audit[AUDIT_setgid]==0) return;
			sys_call_table[SYS_setgid]=orig_setgid;
			active_audit[AUDIT_setgid]=0;
			break;
		case SYS_setregid:
			if(active_audit[AUDIT_setregid]==0) return;
			sys_call_table[SYS_setregid]=orig_setregid;
			active_audit[AUDIT_setregid]=0;
			break;
		case SYS_setresgid:
			if(active_audit[AUDIT_setresgid]==0) return;
			sys_call_table[SYS_setresgid]=orig_setresgid;
			active_audit[AUDIT_setresgid]=0;
			break;
		case SYS_setgid32:
			if(active_audit[AUDIT_setgid32]==0) return;
			sys_call_table[SYS_setgid32]=orig_setgid32;
			active_audit[AUDIT_setgid32]=0;
			break;
		case SYS_setregid32:
			if(active_audit[AUDIT_setregid32]==0) return;
			sys_call_table[SYS_setregid32]=orig_setregid32;
			active_audit[AUDIT_setregid32]=0;
			break;
		case SYS_setresgid32:
			if(active_audit[AUDIT_setresgid32]==0) return;
			sys_call_table[SYS_setresgid32]=orig_setresgid32;
			active_audit[AUDIT_setresgid32]=0;
			break;
		case SYS_truncate64:
			if(active_audit[AUDIT_truncate64]==0) return;
			sys_call_table[SYS_truncate64]=orig_truncate64;
			active_audit[AUDIT_truncate64]=0;
			break;
//		case SYS_ftruncate:
//			if(active_audit[AUDIT_ftruncate]==0) return;
//			sys_call_table[SYS_ftruncate]=orig_ftruncate;
//			active_audit[AUDIT_ftruncate]=0;
//			break;
//		case SYS_fchmod:
//			if(active_audit[AUDIT_fchmod]==0) return;
//			sys_call_table[SYS_fchmod]=orig_fchmod;
//			active_audit[AUDIT_fchmod]=0;
//			break;
//		case SYS_ftruncate64:
//			if(active_audit[AUDIT_ftruncate64]==0) return;
//			sys_call_table[SYS_ftruncate64]=orig_ftruncate64;
//			active_audit[AUDIT_ftruncate64]=0;
//			break;
//		case SYS_fchown:
//			if(active_audit[AUDIT_fchown]==0) return;
//			sys_call_table[SYS_fchown]=orig_fchown;
//			active_audit[AUDIT_fchown]=0;
//			break;
	}
}


////////////////////////////////////////////////////////////////////////////////
// Control the reporting of the audit event.
////////////////////////////////////////////////////////////////////////////////

int audit_event(void * event, int userset)
{
	int class=0;
	static struct timeval time_temp;	// to get the current time
	static null_class * any_event;
	int signal;							// Signal the daemon
	signal=1;

	any_event = event;

	// No use auditing anything unless the auditd is listening.
	if(auditdaemon_task_struct == NULL || audit_device_open == 0)
		return -ENODEV;

	// Reset both counts if we are close to maxint, otherwise we will have strange statistics.
	if(total_events>=(INT_MAX - 1)) {
		total_events=0;
		lost_events=0;
	}
	total_events++;

	// The routine that sends data to us should have filled in
	// the event_id in the header_token.
	// And the returncode should also be known.
	class = any_event->t_header.event_class;

	// fill in the time of the event
	// has to be when locked because otherwise we could interrupt and that may
	// generate an audited event - the times would then be out of order
	// Lock the command execution path until spin_lock_irqrestore
	spin_lock_irqsave(&spin_lock, flags);
	do_gettimeofday(&time_temp);
	// Allow the kernel to continue execution, and handle signals.
	spin_unlock_irqrestore(&spin_lock, flags);

	any_event->t_header.time.tv_sec = time_temp.tv_sec;
	any_event->t_header.time.tv_usec = time_temp.tv_usec;

	// If userid is already set, don't grab it again.
	if(!userset) {
		// Get the current user data
		// This will be the same for any event.
		any_event->t_header.user_id=current->uid;
		any_event->t_header.group_id=current->gid;
		any_event->t_header.euser_id=current->euid;
		any_event->t_header.egroup_id=current->egid;
	}

	if(class == AUDIT_CLASS_IO)
	{
		// do somethin with the data.
		if(!write_io((io_class *)event))
		{
			// Couldnt write it out. Missed this event.
			signal=0;
			lost_events++;
		}
	} else if (class == AUDIT_CLASS_EXEC) {
		if(!write_ex((ex_class *)event))
		{
			// Couldnt write it out. Missed this event.
			signal=0;
			lost_events++;
		}
	} else if (class == AUDIT_CLASS_PC) {
		if(!write_pc((pc_class *)event))
		{
			// Couldnt write it out. Missed this event.
			signal=0;
			lost_events++;
		}
	} else if (class == AUDIT_CLASS_CH) {
		if(!write_ch((ch_class *)event))
		{
			// Couldnt write it out. Missed this event.
			signal=0;
			lost_events++;
		}
	} else if (class == AUDIT_CLASS_CP) {
		if(!write_cp((cp_class *)event))
		{
			// Couldnt write it out. Missed this event.
			signal=0;
			lost_events++;
		}
	} else if (class == AUDIT_CLASS_SU) {
		if(!write_su((su_class *)event))
		{
			// Couldnt write it out. Missed this event.
			signal=0;
			lost_events++;
		}
	} else {
		printk("Cannot handle this class yet.\n");
		signal=0;
		// Add 1 to the lost events tally
		lost_events++;
	}

	if(signal)
		signal_auditd();

	return(1); // no event handler
}

////////////////////////////////////////////////////////////////////////////////
// Safely write out the data to our buffer
//
// This routine should be spin_locked for IRQ because it relies on variables
// that may be changed when the read routine is signaled/executed (which may
// happen half way through this routine without locking)
//	storage_buffer_start, write_position
//
// WARNING: DO NOT CHANGE OR RELY ON THE VARIABLES ABOVE UNLESS YOU ARE
//		  IN A SPIN_LOCKED STATE - THEY MAY BE CORRUPTED BY THE
//		  AUDIT_READ MODULE, WHICH IS CALLED BY SIGNAL/INTERRUPT
//
////////////////////////////////////////////////////////////////////////////////
int write_event(void * event, int token_size)
{
	int bytes_free=0;
	int contiguous_bytes_avail=0;
	int bytes_remaining=0;
	int Blocking;
	
	struct timeval starttime;	

	// Potentially, Lock the kernel.
	// We'll be playing with variables that may change on the
	// fly within audit_read otherwise!

	Blocking=1;

	if(audit_delivery == DELIVERY_TIMEOUT) {
		spin_lock_irqsave(&spin_lock, flags);
		do_gettimeofday(&starttime);
		spin_unlock_irqrestore(&spin_lock, flags);
	}
	
	// Don't allow interrupts to occur while we are here.
	spin_lock_irqsave(&spin_lock, flags);

	while(Blocking)
	{
		// Basics: write out data, respecting our circular buffer.

		// Work out the number of bytes available.
		// Difference between read and write buffers (rotating buffer)
		if(read_position > write_position)
		{
			// Subtract one byte to ensure that the write position never catches back up to the read position.
			// Ensure that the write position never catches back up to the read position.
			// bytes_free=read_position - write_position - 1;
			bytes_free=read_position - write_position;
			contiguous_bytes_avail=bytes_free;
			queue_bytes_free = bytes_free;
		} else {
			// Add the distance between the read position and the end of the buffer to
			// the distance between the start of the buffer and the write position.
			// Subtract one byte to ensure that the write position never catches back up to the read position.
			// (Since we use if(read != write) to determine whether there is any data available.)
			bytes_free=((storage_buffer_start + buffer_size) - write_position) + (read_position - storage_buffer_start);
			contiguous_bytes_avail=(storage_buffer_start + buffer_size) - write_position;
			queue_bytes_free = bytes_free;
		}

		// If we could not write the event, and the user has requested that
		// they wish to force delivery.
		if(bytes_free < token_size)
		{
			// NOTE: TIMEOUT is not yet available as an option.
			if(audit_delivery == DELIVERY_TIMEOUT)
			{
				// Sorry - we cannot write this out, otherwise we would be
				// Corrupting our own read space.
				// ie: The dog would have swallowed his own tail.
			
				// We will have to block until there is space available in the buffer.
				spin_unlock_irqrestore(&spin_lock, flags);
				// queue_bytes_free is a global variable which will change when an event is read by the
				// audit daemon.
				// OR, if the user has requested that a 5 second timeout be set, 
				wait_event(audit_queue, (queue_bytes_free >= token_size || timeout(&starttime)));

				if(timeout(&starttime)) {
					printk("Audit DEBUG: TIMEOUT\n");
					return(0);
				}

				spin_lock_irqsave(&spin_lock, flags);
				// No timeout yet. Go back and try writing again now.
				Blocking=1;
			} else if(audit_delivery == DELIVERY_C2) {
				// The user REALLY wants this event to be delivered.
				// At the POTENTIAL risk of the kernel halting if the audit daemon
				// somehow suspends with the /proc file open!
				spin_unlock_irqrestore(&spin_lock, flags);
				wait_event(audit_queue, (queue_bytes_free >= token_size));
				// Lock ourselves down again.
				spin_lock_irqsave(&spin_lock, flags);
				// Note that we cannot release the lock here, as the variables
				// just below the while(Blocking) loop starting point
				// MUST be reevaluated before going any further.
			} else {
				// release our spin lock
				spin_unlock_irqrestore(&spin_lock, flags);
				// If the user is not interested in guaranteed audit event delivery.
				return(0);
			}
		} else {
			// Excellent, there is enough space in the buffer.
			Blocking=0;
		}
	}

	// Write out the data prior to the end of the buffer.
	if(contiguous_bytes_avail >= token_size)
	{
		// This will always be the case if write position < read position.
		// But MAY not be the case if read position < write position
		memcpy(write_position,event,token_size);
		write_position+=token_size;

		// Make sure we have not rolled past the buffer end point.
		if(write_position >= storage_buffer_start + buffer_size) {
			write_position = storage_buffer_start;
		}
	} else {
		memcpy(write_position,event,contiguous_bytes_avail);
		// reset our write pointer to the start of the buffer
		// and copy the rest of the buffer data
		write_position=storage_buffer_start;
		bytes_remaining=token_size-contiguous_bytes_avail;
		memcpy(write_position,event+contiguous_bytes_avail,bytes_remaining);
		write_position+=bytes_remaining;
	}

	// Allow the kernel to continue execution, and handle signals.
	spin_unlock_irqrestore(&spin_lock, flags);

	return(1);
}

// Determine if a timeout has occurred.
int timeout(struct timeval *start)
{
	struct timeval currenttime;

	spin_lock_irqsave(&spin_lock, flags);
	do_gettimeofday(&currenttime);
	spin_unlock_irqrestore(&spin_lock, flags);
	
	if((currenttime.tv_sec - start->tv_sec) >= 5) {
		return(1);
	} else {
		// Sleep a little before returning.
		queue_task(&Task,&tq_timer);
		sleep_on(&wait_queue);
		return(0);
	}
}

// Sleep for 1000 ticks before sending a wake-up call.
// what if the user rmmods the module while this is still revolving?
// Lets disable the timeout feature until we resolve this.
void timecheck(void * noarg)
{
	static long counter=0;

	counter++;
	if(counter > 1000) {
		counter=0;
		wake_up(&wait_queue);
	} else {
		// Re-queue ourselves.
		queue_task(&Task,&tq_timer);
	}		
}


////////////////////////////////////////////////////////////////////////////////
// The following section contains the broad write processes for each class
// of audit event.
////////////////////////////////////////////////////////////////////////////////

// Write input/output style events out to our buffer.
int write_io(io_class * event)
{
	char tempbuffer[MAX_PATH];
	// Set any other io related stuff

	// Get the current working directory.
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,tempbuffer,MAX_PATH),MAX_PATH);
#else
	// New for kernel 2.4 - check if pwdmnt is appropriate. Alternatively: rootmnt/altrootmnt
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,current->fs->rootmnt,tempbuffer,MAX_PATH),MAX_PATH);
#endif
	// Set the size to io_class minus header_token.
	event->t_header.event_size=sizeof(io_class) - sizeof(header_token);

	return(write_event(event,sizeof(io_class)));
}

// Write input/output chown style events out to our buffer.
int write_ch(ch_class * event)
{
	char tempbuffer[MAX_PATH];
	// Set any other ch related stuff

	// Get the current working directory.
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,tempbuffer,MAX_PATH),MAX_PATH);
#else
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,current->fs->rootmnt,tempbuffer,MAX_PATH),MAX_PATH);
#endif

	// Set the size to ch_class minus header_token.
	event->t_header.event_size=sizeof(ch_class) - sizeof(header_token);

	return(write_event(event,sizeof(ch_class)));
}

// Write execution style events out to our buffer.
int write_ex(ex_class * event)
{
	char tempbuffer[MAX_PATH];

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,tempbuffer,MAX_PATH),MAX_PATH);
#else
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,current->fs->rootmnt,tempbuffer,MAX_PATH),MAX_PATH);
#endif

	// Set the size to ex_class minus header_token.
	event->t_header.event_size=sizeof(ex_class) - sizeof(header_token);

	return(write_event(event,sizeof(ex_class)));
}

// Write process control style events out to our buffer.
int write_pc(pc_class * event)
{
	// Set the size to pc_class minus header_token.
	event->t_header.event_size=sizeof(pc_class) - sizeof(header_token);
	
	return(write_event(event,sizeof(pc_class)));
}

// Write input/output style events with a double path out to our buffer.
int write_cp(cp_class * event)
{
	char tempbuffer[MAX_PATH];

	// Get the current working directory.
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,tempbuffer,MAX_PATH),MAX_PATH);
#else
	strncpy(event->t_pwd.path,d_path(current->fs->pwd,current->fs->rootmnt,tempbuffer,MAX_PATH),MAX_PATH);
#endif
	// Set the size to cp_class minus header_token.
	event->t_header.event_size=sizeof(cp_class) - sizeof(header_token);

	return(write_event(event,sizeof(cp_class)));
}

// Write process control style events out to our buffer.
int write_su(su_class * event)
{
	// Set the size to su_class minus header_token.
	event->t_header.event_size=sizeof(su_class) - sizeof(header_token);
	
	return(write_event(event,sizeof(su_class)));
}


////////////////////////////////////////////////////////////////////////////////
// The following section contains the modified system calls that include
// the auditing that we require.
// Note that the kernel will be IRQ spin-locked once the audit_event routine
// is called.
////////////////////////////////////////////////////////////////////////////////

// Audit file opens.
asmlinkage int audit_open(const char *pathname, int flag, mode_t mode)
{
	io_class	io_event;

	io_event.t_return.returncode=orig_open(pathname, flag, mode);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(io_event.t_return.returncode);
	}

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_open;

	// Process data
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,pathname,MAX_PATH);
	io_event.t_attributes.mode=flag;
	io_event.t_attributes.createmode=mode;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	return(io_event.t_return.returncode);
}

// Audit file opens (create only).
asmlinkage int audit_creat(const char *pathname, mode_t mode)
{
	io_class	io_event;

	io_event.t_return.returncode=orig_creat(pathname, mode);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(io_event.t_return.returncode);
	}

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_creat;

	// Process data
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,pathname,MAX_PATH);
	io_event.t_attributes.mode=O_CREAT|O_WRONLY|O_TRUNC;
	io_event.t_attributes.createmode=mode;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	return(io_event.t_return.returncode);
}

// Audit execve command executions
// Execve is not handled quite the same as other audit events.
// 
int audit_execve(struct pt_regs regs)
{
	ex_class	ex_event;

	char *filename=NULL;
	char **argv=NULL;

	unsigned int delimiter_required=0;

	// Not really sure I need to getname here.
	// After-all, I'm not sending this to user-space yet.
	// I'll keep it in for safety.
	filename = getname((char*)regs.ebx);
	argv = (char **)regs.ecx;

	// If audit is not active, don't go any further.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		ex_event.t_return.returncode=do_execve(filename, (char**)regs.ecx, (char**)regs.edx, &regs);
		putname(filename);
		return(ex_event.t_return.returncode);
	}

	// Construct the argument list
	// and shoehorn it into exec_args

	ex_event.t_execargs.args[0]='\0';
	while(*argv != NULL)
	{
		// Allow for the delimiter
		if((strlen(ex_event.t_execargs.args) + strlen(*argv)) < MAX_PATH)
		{
			if(!delimiter_required)
			{
				delimiter_required=1;
			} else {
				strcat(ex_event.t_execargs.args," ");
			}
			strcat(ex_event.t_execargs.args,*argv);
		} else {
			break;
		}
		argv++;
	}

	// Process data
	ex_event.t_process.pid=current->pid;
	ex_event.t_process.name[0]='\0';
	strncpy(ex_event.t_process.name,current->comm,MAXCOMMAND);
	
	ex_event.t_return.returncode = do_execve(filename, (char**)regs.ecx, (char**)regs.edx, &regs);
	ex_event.t_path.path[0]='\0';
	strncpy(ex_event.t_path.path,filename,MAX_PATH);

	putname(filename);

	ex_event.t_header.event_class=AUDIT_CLASS_EXEC;
	ex_event.t_header.event_id=SYS_execve;

	// Everything else can be handled by the generic event handler.
	audit_event(&ex_event,0);

	return(ex_event.t_return.returncode);
}


// Audit process exits.
// This may be useful if we want to audit login/logouts - because
// as far as I can work out, the only way we can do it is to watch for
// execve/exit of /bin/login (without hacking the kernel).
//
// Note: current->comm is a 16 character representation of the executed command.
asmlinkage int audit_exit(int status)
{
	pc_class	pc_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_exit(status);
	}

	pc_event.t_return.returncode=status;

	pc_event.t_header.event_class=AUDIT_CLASS_PC;
	pc_event.t_header.event_id=SYS_exit;

	pc_event.t_process.pid=current->pid;
	// Zero the string just in case this fails.
	pc_event.t_process.name[0]='\0';
	strncpy(pc_event.t_process.name,current->comm,MAXCOMMAND);

	// Everything else can be handled by the generic event handler.
	audit_event(&pc_event,0);

	// Now continue on with the normal exit.
	return orig_exit(status);
}

// Audit mkdir directory creations.
asmlinkage int audit_mkdir(const char *path, mode_t mode)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_mkdir(path,mode);
	}

	io_event.t_return.returncode=orig_mkdir(path,mode);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_mkdir;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	io_event.t_attributes.mode=0;
	io_event.t_attributes.createmode=(unsigned long)mode;

	// Store the path
	strncpy(io_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// audit file unlinks
asmlinkage int	audit_unlink (const char *path)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_unlink(path);
	}

	io_event.t_return.returncode=orig_unlink(path);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_unlink;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Store the path
	// zero the path, just in case strncpy fails.
	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// Create a special file/directory
asmlinkage int audit_mknod (const char *pathname, mode_t mode, dev_t dev)
{
	io_class	io_event;

	io_event.t_return.returncode=orig_mknod(pathname, mode, dev);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(io_event.t_return.returncode);
	}

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_mknod;

	// Process data
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,pathname,MAX_PATH);
	io_event.t_attributes.mode=(int)mode;
	io_event.t_attributes.createmode=(unsigned long)dev;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	return(io_event.t_return.returncode);
}

// Audit rmdir directory removals.
asmlinkage int audit_rmdir(const char *path)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_rmdir(path);
	}

	io_event.t_return.returncode=orig_rmdir(path);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_rmdir;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Store the path
	strncpy(io_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);
	
	return io_event.t_return.returncode;
}

// Audit change ownership
asmlinkage int audit_chown(const char *path, uid_t owner, gid_t group)
{
	ch_class	ch_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_chown(path,owner,group);
	}

	ch_event.t_return.returncode=orig_chown(path,owner,group);

	ch_event.t_header.event_class=AUDIT_CLASS_CH;
	ch_event.t_header.event_id=SYS_chown;
	ch_event.t_process.pid=current->pid;
	ch_event.t_process.name[0]='\0';
	strncpy(ch_event.t_process.name,current->comm,MAXCOMMAND);

	ch_event.t_owner.owner=(int) owner;
	ch_event.t_owner.group=(int) group;

	// Store the path
	strncpy(ch_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&ch_event,0);

	// Now continue on with the normal exit.
	return ch_event.t_return.returncode;

}

// Same as chown, but does not follow symlinks.
asmlinkage int audit_lchown(const char *path, uid_t owner, gid_t group)
{
	ch_class	ch_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_lchown(path,owner,group);
	}

	ch_event.t_return.returncode=orig_lchown(path,owner,group);

	ch_event.t_header.event_class=AUDIT_CLASS_CH;
	ch_event.t_header.event_id=SYS_lchown;
	ch_event.t_process.pid=current->pid;
	ch_event.t_process.name[0]='\0';
	strncpy(ch_event.t_process.name,current->comm,MAXCOMMAND);

	ch_event.t_owner.owner=(int) owner;
	ch_event.t_owner.group=(int) group;

	// Store the path
	strncpy(ch_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&ch_event,0);

	// Now continue on with the normal exit.
	return ch_event.t_return.returncode;
}

// Change ownership
asmlinkage int audit_chown32(const char *path, uid_t owner, gid_t group)
{
	ch_class	ch_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_chown32(path,owner,group);
	}

	ch_event.t_return.returncode=orig_chown32(path,owner,group);

	ch_event.t_header.event_class=AUDIT_CLASS_CH;
	ch_event.t_header.event_id=SYS_chown32;
	ch_event.t_process.pid=current->pid;
	ch_event.t_process.name[0]='\0';
	strncpy(ch_event.t_process.name,current->comm,MAXCOMMAND);

	ch_event.t_owner.owner=(int) owner;
	ch_event.t_owner.group=(int) group;

	// Store the path
	strncpy(ch_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&ch_event,0);

	// Now continue on with the normal exit.
	return ch_event.t_return.returncode;

}

// Same as chown, but does not follow symlinks.
asmlinkage int audit_lchown32(const char *path, uid_t owner, gid_t group)
{
	ch_class	ch_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_lchown32(path,owner,group);
	}

	ch_event.t_return.returncode=orig_lchown32(path,owner,group);

	ch_event.t_header.event_class=AUDIT_CLASS_CH;
	ch_event.t_header.event_id=SYS_lchown32;
	ch_event.t_process.pid=current->pid;
	ch_event.t_process.name[0]='\0';
	strncpy(ch_event.t_process.name,current->comm,MAXCOMMAND);

	ch_event.t_owner.owner=(int) owner;
	ch_event.t_owner.group=(int) group;

	// Store the path
	strncpy(ch_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&ch_event,0);

	// Now continue on with the normal exit.
	return ch_event.t_return.returncode;

}


// Change the mode of a file or directory
asmlinkage int audit_chmod (const char *pathname, mode_t mode)
{
	io_class	io_event;

	io_event.t_return.returncode=orig_chmod(pathname, mode);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(io_event.t_return.returncode);
	}

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_chmod;

	// Process data
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,pathname,MAX_PATH);
	io_event.t_attributes.mode=(int)NULL;
	io_event.t_attributes.createmode=(unsigned long)mode;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	return(io_event.t_return.returncode);
}

// Create a symbolic link
asmlinkage int audit_symlink (const char *oldpath, const char *newpath)
{
	cp_class	cp_event;

	cp_event.t_return.returncode=orig_symlink(oldpath, newpath);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(cp_event.t_return.returncode);
	}

	cp_event.t_header.event_class=AUDIT_CLASS_CP;
	cp_event.t_header.event_id=SYS_symlink;

	// Process data
	cp_event.t_process.pid=current->pid;
	cp_event.t_process.name[0]='\0';
	strncpy(cp_event.t_process.name,current->comm,MAXCOMMAND);

	cp_event.t_sourcepath.path[0]='\0';
	strncpy(cp_event.t_sourcepath.path,oldpath,MAX_PATH);
	cp_event.t_destpath.path[0]='\0';
	strncpy(cp_event.t_destpath.path,newpath,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&cp_event,0);

	return(cp_event.t_return.returncode);
}

// Create a hard link
asmlinkage int audit_link (const char *oldpath, const char *newpath)
{
	cp_class	cp_event;

	cp_event.t_return.returncode=orig_link(oldpath, newpath);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(cp_event.t_return.returncode);
	}

	cp_event.t_header.event_class=AUDIT_CLASS_CP;
	cp_event.t_header.event_id=SYS_link;

	// Process data
	cp_event.t_process.pid=current->pid;
	cp_event.t_process.name[0]='\0';
	strncpy(cp_event.t_process.name,current->comm,MAXCOMMAND);

	cp_event.t_sourcepath.path[0]='\0';
	strncpy(cp_event.t_sourcepath.path,oldpath,MAX_PATH);
	cp_event.t_destpath.path[0]='\0';
	strncpy(cp_event.t_destpath.path,newpath,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&cp_event,0);

	return(cp_event.t_return.returncode);
}

// rename a file/directory
asmlinkage int audit_rename (const char *oldpath, const char *newpath)
{
	cp_class	cp_event;

	cp_event.t_return.returncode=orig_rename(oldpath, newpath);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(cp_event.t_return.returncode);
	}

	cp_event.t_header.event_class=AUDIT_CLASS_CP;
	cp_event.t_header.event_id=SYS_rename;

	// Process data
	cp_event.t_process.pid=current->pid;
	cp_event.t_process.name[0]='\0';
	strncpy(cp_event.t_process.name,current->comm,MAXCOMMAND);

	cp_event.t_sourcepath.path[0]='\0';
	strncpy(cp_event.t_sourcepath.path,oldpath,MAX_PATH);
	cp_event.t_destpath.path[0]='\0';
	strncpy(cp_event.t_destpath.path,newpath,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&cp_event,0);

	return(cp_event.t_return.returncode);
}

// Would like to audit reboots here.
// NOTE: this only audits when the reboot() system call is issued. There are many other
//       ways to reboot a system (eg: kill -9 1) - this will NOT catch these.
//       In fact, a normal reboot from the gdm login manager (for example), does not
//       seem to call the reboot() call. So this call may not be particularly useful.
// Only question is: does reboot return only after the system goes down?
// (in which case, I'll have to do the audit, fake the return code, and
// return with the orig_reboot().
// After all... we don't want to try and write stuff after filesystems have been unmounted.
asmlinkage int audit_reboot(int magic, int magic2, int flag, void *arg)
{
	pc_class	pc_event;

	pc_event.t_return.returncode=orig_reboot(magic, magic2, flag, arg);

	// If no daemon, or the daemon has generated the event.
	// Or if the reboot flags are just disabling or reenabling the CAD key
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)) ||
	flag == LINUX_REBOOT_CMD_CAD_ON || flag == LINUX_REBOOT_CMD_CAD_OFF)
	{
		return(pc_event.t_return.returncode);
	}

	pc_event.t_header.event_class=AUDIT_CLASS_PC;
	pc_event.t_header.event_id=SYS_reboot;

	// Process data
	pc_event.t_process.pid=current->pid;
	pc_event.t_process.name[0]='\0';
	strncpy(pc_event.t_process.name,current->comm,MAXCOMMAND);

	// Everything else can be handled by the generic event handler.
	audit_event(&pc_event,0);

	return(pc_event.t_return.returncode);
}

// audit file truncates.
// Note that off_t can be shoehorned into a unsigned long.
asmlinkage int	audit_truncate (const char *path, off_t length)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_truncate(path,length);
	}

	io_event.t_return.returncode=orig_truncate(path,length);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_truncate;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Store the path
	// zero the path, just in case strncpy fails.
	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,path,MAX_PATH);
	
	// Save off the offset.
	io_event.t_attributes.createmode=(unsigned long)length;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// audit root changes.
asmlinkage int	audit_chroot (const char *path)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_chroot(path);
	}

	io_event.t_return.returncode=orig_chroot(path);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_chroot;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Store the path
	// zero the path, just in case strncpy fails.
	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,path,MAX_PATH);

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// Setuid system calls
asmlinkage int	audit_setuid (uid_t uid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setuid(uid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setuid(uid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setuid;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=uid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setreuid (uid_t ruid, uid_t euid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setreuid(ruid,euid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setreuid(ruid,euid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setreuid;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=euid;
	su_event.t_target.rid=ruid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

// Set real, effective and saved user ID
asmlinkage int	audit_setresuid (uid_t ruid, uid_t euid, uid_t suid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setresuid(ruid,euid,suid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setresuid(ruid,euid,suid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setresuid;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=euid;
	su_event.t_target.rid=ruid;
	su_event.t_target.sid=suid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setuid32 (uid_t uid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setuid32(uid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setuid32(uid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setuid32;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=uid;
	su_event.t_target.rid=0;
	su_event.t_target.sid=0;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setreuid32 (uid_t ruid, uid_t euid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setreuid32(ruid,euid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setreuid32(ruid,euid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setreuid32;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=euid;
	su_event.t_target.rid=ruid;
	su_event.t_target.sid=0;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setresuid32 (uid_t ruid, uid_t euid, uid_t suid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setresuid32(ruid,euid,suid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setresuid32(ruid,euid,suid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setresuid32;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=euid;
	su_event.t_target.rid=ruid;
	su_event.t_target.sid=suid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setgid (gid_t gid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setgid(gid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setgid(gid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setgid;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=gid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setregid (gid_t rgid, gid_t egid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setregid(rgid,egid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setregid(rgid,egid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setregid;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=egid;
	su_event.t_target.rid=rgid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setresgid (gid_t rgid, gid_t egid, gid_t sgid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setresgid(rgid,egid,sgid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setresgid(rgid,egid,sgid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setresgid;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=egid;
	su_event.t_target.rid=rgid;
	su_event.t_target.sid=sgid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setgid32 (gid_t gid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setgid32(gid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setgid32(gid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setgid32;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=gid;
	su_event.t_target.rid=0;
	su_event.t_target.sid=0;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setregid32 (gid_t rgid, gid_t egid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setregid32(rgid,egid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setregid32(rgid,egid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setregid32;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=egid;
	su_event.t_target.rid=rgid;
	su_event.t_target.sid=0;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}

asmlinkage int	audit_setresgid32 (gid_t rgid, gid_t egid, gid_t sgid)
{
	su_class	su_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_setresgid32(rgid,egid,sgid);
	}

	// Set the current id data from the pre-system call user data
	su_event.t_header.user_id=current->uid;
	su_event.t_header.group_id=current->gid;
	su_event.t_header.euser_id=current->euid;
	su_event.t_header.egroup_id=current->egid;

	su_event.t_return.returncode=orig_setresgid32(rgid,egid,sgid);

	su_event.t_header.event_class=AUDIT_CLASS_SU;
	su_event.t_header.event_id=SYS_setresgid32;
	su_event.t_process.pid=current->pid;
	su_event.t_process.name[0]='\0';
	strncpy(su_event.t_process.name,current->comm,MAXCOMMAND);

	su_event.t_target.id=egid;
	su_event.t_target.rid=rgid;
	su_event.t_target.sid=sgid;

	// Everything else can be handled by the generic event handler.
	audit_event(&su_event,1);

	// Now continue on with the normal exit.
	return su_event.t_return.returncode;
}


// audit file truncates.
// NOTE: This event is never called! Linux works out the filename, and calls truncate() instead!
asmlinkage int	audit_ftruncate (int fd, off_t length)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_ftruncate(fd,length);
	}

	io_event.t_return.returncode=orig_ftruncate(fd,length);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_ftruncate;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Ftruncate has no path, just a file int.
	// zero the path, just in case strncpy fails.
	io_event.t_path.path[0]='\0';
	// Would prefer to use snprintf here, but it's not available within the module
	// As long as MAX_PATH has enough storage for the text representation of an
	// integer, we'll be fine (ie: no buffer overflow).
	sprintf(io_event.t_path.path,"FD=%d",fd);
	
	// Save off the offset.
	io_event.t_attributes.createmode=(unsigned long)length;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// NOTE: This event is never called! Linux works out the filename, and calls chmod() instead!
asmlinkage int audit_fchmod (int filedes, mode_t mode)
{
	io_class	io_event;

	io_event.t_return.returncode=orig_fchmod(filedes, mode);

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return(io_event.t_return.returncode);
	}


	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_chmod;

	// Process data
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	io_event.t_path.path[0]='\0';
	sprintf(io_event.t_path.path,"FD=%d",filedes);

	io_event.t_attributes.mode=(int)NULL;
	io_event.t_attributes.createmode=(unsigned long)mode;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	return(io_event.t_return.returncode);
}


// audit file truncates.
// NOTE: This event is never called! Linux works out the filename, and calls truncate64() instead!
asmlinkage int	audit_ftruncate64 (int fd, off_t length)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_ftruncate64(fd,length);
	}

	io_event.t_return.returncode=orig_ftruncate64(fd,length);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_ftruncate64;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Ftruncate has no path, just a file int.
	// zero the path, just in case strncpy fails.
	io_event.t_path.path[0]='\0';
	sprintf(io_event.t_path.path,"FD=%d",fd);
	
	// Save off the offset.
	io_event.t_attributes.createmode=(unsigned long)length;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// audit file truncates.
asmlinkage int	audit_truncate64 (const char *path, off_t length)
{
	io_class	io_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_truncate64(path,length);
	}

	io_event.t_return.returncode=orig_truncate64(path,length);

	io_event.t_header.event_class=AUDIT_CLASS_IO;
	io_event.t_header.event_id=SYS_truncate64;
	io_event.t_process.pid=current->pid;
	io_event.t_process.name[0]='\0';
	strncpy(io_event.t_process.name,current->comm,MAXCOMMAND);

	// Store the path
	// zero the path, just in case strncpy fails.
	io_event.t_path.path[0]='\0';
	strncpy(io_event.t_path.path,path,MAX_PATH);
	
	// Save off the offset.
	io_event.t_attributes.createmode=(unsigned long)length;

	// Everything else can be handled by the generic event handler.
	audit_event(&io_event,0);

	// Now continue on with the normal exit.
	return io_event.t_return.returncode;
}

// fchown is not actually called! Linux calls chown instead (chown32)!
asmlinkage int audit_fchown(int fd, uid_t owner, gid_t group)
{
	// Dunno whether we should use an io class for chown.
	ch_class	ch_event;

	// If no daemon, or the daemon has generated the event.
	if((!AUDIT_IS_RUNNING) || (AUDIT_IS_RUNNING && (current->pid == auditdaemon_task_struct->pid)))
	{
		return orig_fchown(fd,owner,group);
	}

	ch_event.t_return.returncode=orig_fchown(fd,owner,group);

	ch_event.t_header.event_class=AUDIT_CLASS_CH;
	ch_event.t_header.event_id=SYS_fchown;
	ch_event.t_process.pid=current->pid;
	ch_event.t_process.name[0]='\0';
	strncpy(ch_event.t_process.name,current->comm,MAXCOMMAND);

	ch_event.t_owner.owner=(int) owner;
	ch_event.t_owner.group=(int) group;

	// Fchown has no path, just a file int.
	// zero the path, just in case strncpy fails.
	ch_event.t_path.path[0]='\0';
	sprintf(ch_event.t_path.path,"FD=%d",fd);

	// Everything else can be handled by the generic event handler.
	audit_event(&ch_event,0);

	// Now continue on with the normal exit.
	return ch_event.t_return.returncode;

}
